/*
Qhull enviada por Coro desde Edimburgo.10.98
*/

#define CCMPI
#ifdef CCPVM
#define BARRIER pvm_barrier("PVM_COMM_WORLD",NUMPROCESSORS)
#define GROUP	pvm_joingroup("PVM_COMM_WORLD")
#else
#define BARRIER MPI_Barrier(MPI_COMM_WORLD)
#define GROUP	
#endif


/* ----- Libraries ----- */
#include "../llcsync.h"
#include "../llcprnt.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <mytime.h>


/* ----- Constants ----- */
/* #define OUTFILE 1 */
#define Epsilon 1.2E-7

#define Mb  3
#define K    1024
#define R    Mb *  K * K
#define R2   (doubleType)((doubleType)(R)*(doubleType)(R))
#define SIZE (2 * R) + 1

#define MAX_TIMES 10
#define MAX_SIZE  1            /* maximun size of the problem N Mb */

#define MAX_DEEP   20
#define MAX_REP 1	



/*					 NUEVO				*/
/******* chronos	 */

CLOCKTYPE Timediv1,Timediv2,FinalTime,InitialTime,TimeS1,TimeS2,TimeP1,TimeP2;
CLOCKTYPE com_time1,com_time2, seqTime1, seqTime2;
CLOCKTYPE combinationTime1,  combinationTime2;

double t_fst[MAX_TIMES]; /* only one division M-step in BSPWB */
double t_division[MAX_TIMES][MAX_DEEP];   /* there are as many as recursion depth */
double t_combination[MAX_TIMES][MAX_DEEP]; /* there are as many as recursion depth */
double t_sequential[MAX_TIMES];
double t_comm[MAX_TIMES][MAX_DEEP];
int sizeenv[MAX_TIMES][MAX_DEEP], sizerec[MAX_TIMES][MAX_DEEP];
double t_seq[MAX_REP], t_par[MAX_REP];  

/******* variables globales	 */

int REP;
int deep1=0;
int deep=1;deep2=1;

#define PARHP(f1, v1, s1, f2, v2, s2)							\
{											\
  bit--;										\
  NUMPROCESSORS >>= 1;									\
  if ((LLNAME & (1 << bit)) == 0) {							\
     f1;										\
		 /*\
		 BARRIER;\
		 */\
		 GETTIME(com_time1); \
     MPI_Sendrecv(&s1, 1, MPI_INT, LLPARTNER(bit), LLCTAG, 				\
                  &s2, 1, MPI_INT, LLPARTNER(bit), LLCTAG, LLC_COMM, &status);	        \
     MPI_Sendrecv(v1, s1, MPI_CHAR, LLPARTNER(bit), LLCTAG, 				\
                  v2, s2, MPI_CHAR, LLPARTNER(bit), LLCTAG, LLC_COMM, &status);	        \
		 GETTIME(com_time2);\
		 /*\
		 BARRIER;\
		 */\
		 GPRINTF("f1--NAME: %d deep= %d  env(s1)= %d rec(s2)= %d \n",LLNAME, deep2, s1,s2);\
		 DIFTIME(com_time2,com_time1,t_comm[REP][deep2]);\
	         sizeenv[REP][deep2]=s1;sizerec[REP][deep2]=s2;\
  }											\
  else {										\
     f2;										\
		 /*\
		 BARRIER;\
		 */\
		 GETTIME(com_time1);\
     MPI_Sendrecv(&s2, 1, MPI_INT, LLPARTNER(bit), LLCTAG, 				\
                  &s1, 1, MPI_INT, LLPARTNER(bit), LLCTAG, LLC_COMM, &status);	        \
     MPI_Sendrecv(v2, s2, MPI_CHAR, LLPARTNER(bit), LLCTAG, 				\
                  v1, s1, MPI_CHAR, LLPARTNER(bit), LLCTAG, LLC_COMM, &status);	        \
		 GETTIME(com_time2);\
		 /*\
		 BARRIER;\
		 */\
		  GPRINTF("f2--NAME: %d  deep= %d  env(s2)= %d rec(s1)= %d \n",LLNAME, deep2, s2,s1);\
 	          DIFTIME(com_time2,com_time1,t_comm[REP][deep2]);\
 		 sizeenv[REP][deep2]=s2;sizerec[REP][deep2]=s1;\
  }                                                                                     \
  bit++;				                                               	\
  NUMPROCESSORS <<= 1;                                                                  \
}

/* ----- Macros ----- */
#define line( A, B, C, p0_x, p0_y, p1_x, p1_y);\
		{ doubleType t0_x, t1_x; \
		t0_x = (doubleType)(p0_x); \
		t1_x = (doubleType)(p1_x); \
		A = p1_y - p0_y;\
		B = t0_x - t1_x;\
		C = - (p0_y) * B - A * t0_x;\
		}

#define signTest( s, A, B, C, t_x, t_y, a);\
		a = (A * (doubleType)(t_x) + B * t_y + C);\
		s = ((a<-(Epsilon)) ? NEG : ( (a>Epsilon) ? POS : NUL));

#define distTest( s, A, B, C, t_x, t_y, signAux, distAux);\
		distAux = (A * (doubleType)(t_x) + B * t_y + C);\
		signAux = distAux;\
		distAux = fabs(distAux); \
		s = ((signAux<-(Epsilon)) ? NEG : ( (signAux>Epsilon) ? POS : NUL));\

/* ----- Types ----- */

typedef short int intType;
typedef long double doubleType;

typedef intType coord_x;	   	/* type for the coordinates of points */
typedef doubleType coord_y;

typedef enum {NEG = -1, NUL = 0, POS = 1} sign;

/* ----- Global Variables  -----*/

coord_x p_x[SIZE];     /* Global array with the x values of the points */
intType hull[SIZE];	 /* Global array with the index of the points */

/* ----- prototypes ----- */
intType generate(void);
void seq_q_hull( intType **first, intType **last, intType *max, intType * size);
void par_q_hull( intType **f_hull, intType **l_hull, intType *m, intType * size);
void show_output( intType *first, intType *last, intType size);

/* ----- Global variables ----- */

int s;                                       /* index loop for sizes */

/*
 * generating the y values of points
 */


 /*  ES LA MISMA QUE ENVIASTE */
coord_y p_y( coord_x x)
{
  coord_y f, t1, t2, t3;

  t1 = (coord_y)(x);
  t2 = t1 * t1;
  t3 = R2 - t2;
  f = sqrt(t3);
  return f;
} /* p_y */


/*
 * generating the x values of points
 */
intType generate() {
  intType i, k=0;

  for (i = -R; i <= R; i++) {
    p_x[k] = i;
    hull[k] = k;
    k++;
  }
  return k;
} /* generate */


/*
 * Select the point with maximun distance to line p[i] - p[j]
 */
intType * maxDistancePoint(intType *i, intType *j)
{
  coord_x pi_x, pj_x, pk_x;       /* Coord of the triangle */
  coord_y pi_y, pj_y, pk_y;
  intType *max = NULL, *k = NULL;
  doubleType dist, dmax;      /* Auxiliary for maximum distance */
  doubleType A, B, C;         /* Parameters of the line canonic equation */

  pi_x = p_x[ *i ];
  pi_y = p_y(pi_x);
  pj_x = p_x[ *j ];
  pj_y = p_y(pj_x);

  /* Line pi-pj */
  line( A, B, C, pi_x, pi_y, pj_x, pj_y);

  /* Selects the point with maximum distance to line pi-pj */
  dmax = 0.0;
  max = 0;
  for (k = i; k <= j; k++) {
    pk_x = p_x[*k];
    pk_y = p_y(pk_x);
    dist = fabs(A * (doubleType)(pk_x) + B * (pk_y) + C);
    if (dist >= dmax) {
      dmax = dist;
      max = k;
    } /* if */
  } /* for */
  if (dmax == 0.0) return NULL;
  else return max;
} /*maxDistacePoint*/

/*
 * Partition:
 *   Puts in the first place of index array the points in the
	 oposite semiplane to p1 with regard to the line p0 - p2
 *   Puts in the last place of index array the points in the
	 oposite semiplane to p0 with regard to the line p1 - p2
 */
void partition( intType *first, intType *last, intType *max, intType **lastA, intType **firstB, intType **maxA, intType **maxB)
{
  sign sig, signA, signB;                     /* Auxiliary for finding signs */
  intType *left, *right, *i, *j, temp;        /* indexs */
  doubleType sigAux, distAux;
  doubleType mv_A = 0.0, mv_B = 0.0;
  intType *max_A = NULL, *max_B = NULL;
  doubleType A1 = 0.0, B1 = 0.0, C1 = 0.0, A2 = 0.0, B2 = 0.0, C2 = 0.0;      /* Parameters of the line canonic equation */
  coord_x pfirst_x, pmax_x, plast_x, pi_x, pj_x;
  coord_y pfirst_y, pmax_y, plast_y, pi_y, pj_y;


  pfirst_x = p_x[*first];	pfirst_y = p_y(pfirst_x);
  pmax_x = p_x[*max];    	pmax_y = p_y(pmax_x);
  plast_x = p_x[*last];  	plast_y = p_y(plast_x);

  /* Line pfirst-pmax */
  line( A1, B1, C1, pfirst_x, pfirst_y, pmax_x, pmax_y);

  signTest( signB, A1, B1, C1, plast_x, plast_y, sigAux);

  /* Line plast-pmax */
  line( A2, B2, C2, pmax_x, pmax_y, plast_x, plast_y);

  signTest( signA, A2, B2, C2, pfirst_x, pfirst_y, sigAux);

  /* Put the max point in the first place */
  left = first + 1;
  temp = *left;
  *left = *max;
  *max = temp;

  i = left + 1;
  right = last;
  j = last - 1;

  pi_x = p_x[*i];
  pi_y = p_y(pi_x);
  pj_x = p_x[*j];
  pj_y = p_y(pj_x);

  while (i <= j)
  {
  /* if p[i] don't belong to B */

    signTest( sig, A2, B2, C2, pi_x, pi_y, sigAux);
    while  (sig == signA)  {
      distTest( sig, A1, B1, C1, pi_x, pi_y, sigAux, distAux);
      if (sig != signB)
      {
        *(++left) = *i;
        if (distAux > mv_A + Epsilon)
        {
          mv_A = distAux;
          max_A = left;
        }
      }
      i++;
      pi_x = p_x[*i];
      pi_y = p_y(pi_x);
      signTest( sig, A2, B2, C2, pi_x, pi_y, sigAux);
    } /* sig == signA */

    signTest( sig, A1, B1, C1, pj_x, pj_y, sigAux);
    while  (sig == signB) {
      distTest( sig, A2, B2, C2, pj_x, pj_y, sigAux, distAux);
      if (sig != signB)
      {
        *(--right) = *j;
        if (distAux > mv_B + Epsilon)
        {
          mv_B = distAux;
          max_B = right;
        }
      }
      j--;
      pj_x = p_x[*j];
      pj_y = p_y(pj_x);
      signTest( sig, A1, B1, C1, pj_x, pj_y, sigAux);
    } /* sig == signA */

    if (i < j)
    {
      temp = *i;
      *(++left) = *j;
      *(--right) = temp;
      i++; j--;

      pi_x = p_x[*i];
      pi_y = p_y(pi_x);
      pj_x = p_x[*j];
      pj_y = p_y(pj_x);
    }
  } /* while (i < j ) */

  /* Put the max point in :
   *      the last place of the A zone and in
   *      the first place of the B zone
   */
  if (max_A == left)
    max_A = (first+1);
  temp = *left;
  *left = *(first + 1);
  *(first + 1) = temp;

  if (left < (right-1))              /* only when lastA != firstB */
    *(right-1) = *(left);

  *(lastA) = left;
  *(firstB) = right-1;

  *(maxA) = ((mv_A == 0.0)? NULL:max_A);
  *(maxB) = ((mv_B == 0.0)? NULL:max_B);
} /*partition*/

void seq_q_hull(intType **f_hull, intType **l_hull, intType *m, intType * size)
{
  intType i, N;
  intType *firstA, *lastA = NULL, *firstB = NULL, *lastB;
  intType *max, *maxA = NULL, *maxB = NULL;
  intType *first, *last;
  intType sizeA = 0, sizeB = 0;


  first = *f_hull;
  last = *l_hull;
  max = m;

  firstA = first;
  lastB = last;

  N = lastB - firstA + 1;
  if ( N <= 2 ){
    first = firstA;
    last = lastB;
  }
  else if (max != NULL)
  {
    partition( firstA, lastB, max, &lastA, &firstB, &maxA, &maxB);
    seq_q_hull( &firstA, &lastA, maxA, &sizeA);
    seq_q_hull( &firstB, &lastB, maxB, &sizeB);

    /* (lastB - firstB) because max is in both zones A and B.*/
    /* The index in hull[firstB] is always max, then we start in firstB+1 */
    for (i = 0; i < (lastB - firstB) ; i++)
      *(lastA + 1 + i) = *(firstB + 1 + i);

    first = firstA;
    last = lastA + (lastB - firstB) ;
  }
  else
  {
    first = firstA;
    last = lastB;
  }

  *(f_hull) = first;
  *(l_hull) = last;
  *(size) = last - first + 1;
} /* seq_q_hull*/
 
#define TYPE int

void par_q_hull(intType **f_hull, intType **l_hull, intType *m, intType * size)
{
 intType i, N;
  intType *firstA, *lastA = NULL, *firstB = NULL, *lastB;
  intType *max, *maxA = NULL, *maxB = NULL;
  intType *first, *last;
  intType sizeA, sizeB;

  if (NUMPROCESSORS == 1) {
		GETTIME(seqTime1);
        	seq_q_hull(f_hull, l_hull, m, size);
 		GETTIME(seqTime2);

		DIFTIME(seqTime2,seqTime1,t_sequential[REP]);
		DIFTIME(seqTime2,TimeP1,t_fst[REP]);

    
  }
  else {
    GETTIME(Timediv1);

    first = *f_hull;
    last = *l_hull;
    max = m;

    firstA = first;
    lastB = last;

    N = lastB - firstA + 1;
    if ( N == 2 ){
      first = firstA;
      last = lastB;
    }
    else if (max != NULL){
      partition( firstA, lastB, max, &lastA, &firstB, &maxA, &maxB);
	GETTIME(Timediv2);   
	DIFTIME(Timediv2,Timediv1,t_division[REP][deep1++]);

      PARHP( par_q_hull( &firstA, &lastA, maxA, &sizeA), firstA, sizeA, 
            par_q_hull( &firstB, &lastB, maxB, &sizeB), firstB, sizeB)
   deep2--;
      /* (lastB - firstB) because max is in both zones A and B.*/
      /* The index in hull[firstB] is always max, then we start in firstB+1 */

	GETTIME(combinationTime1);
      for (i = 0; i < (sizeB-1) ; i++)
        *(lastA + 1 + i) = *(firstB + 1 + i);

      first = firstA;
      last = lastA + (sizeB-1);
	GETTIME(combinationTime2);
	DIFTIME(combinationTime2,combinationTime1,t_combination[REP][deep]);
	deep--;

    }
    else {
      first = firstA;
      last = lastB;
    }

    *(f_hull) = first;
    *(l_hull) = last;
    *(size) = (last - first + 1);
  }

} /* par_q_hull */ 



/*
 * print output
 */
void show_output(intType *first, intType *last, intType size) {
#ifdef OUTFILE
  FILE *out;
#endif
  register intType *i;
  coord_x pi_x;
  coord_y pi_y;

#ifdef OUTFILE
  out= fopen("out", "w");
#endif

  for( i = first; i <= last; i++)
  {
    pi_x = p_x[*i];
    pi_y = p_y(pi_x);
#ifdef OUTFILE
   fprintf(out, "%d \t (%d, %Lg)\n", *i , pi_x, pi_y);
#else
   PRINTF("%d \t (%d, %lf)\n", *i , pi_x, pi_y);
#endif
  }

  PRINTF("size = %d\n", size);

  PRINTF("\n====================\n");
#ifdef OUTFILE
  fclose(out);
#endif
} /* output */


main(int argc, char * argv[]) {
  intType i;	/* Index of the points in the Convex Hull */
  intType *first, *last, *max;
  intType size;
 	int name,d; 						/* Counter */
  
	
	GETTIME(InitialTime);  
/*deep=0;deep2=0;
  deep = ((int)(log(NUMPROCESSORS)/log(2)))-1;
  deep2 =((int)(log(NUMPROCESSORS)/log(2)))-1; 
	printf(" %d, %f - 2- %d, %f \n", deep,deep,deep2,deep2); 
	REP=0;*/
  INITIALIZE(argc,argv);
    deep = ((int)(log(NUMPROCESSORS)/log(2)))-1;
    deep2 =((int)(log(NUMPROCESSORS)/log(2)))-1;
    printf(" %d, %f - 2- %d, %f \n", deep,deep,deep2,deep2);
REP=0; 
  /* ----- Sequential Code ----- */
  
  for(i = 0; i < SIZE; i++)
    hull[i] = 0;
  
  size = generate(); 	/* Generates SIZE points */
  
  first = &hull[0];
  last = &hull[size - 1];

  /* Point with maximum distance to first-last*/
  max = maxDistancePoint( first, last);

  BARRIER;
  GETTIME(TimeS1); 
  seq_q_hull( &first, &last, max, &size); 
  GETTIME(TimeS2); 
  DIFTIME(TimeS2,TimeS1,t_seq[REP]);

  /* show_output(first, last, size); */ 

  /* ----- Parallel Code ----- */
  for(i = 0; i < SIZE; i++)
    hull[i] = 0;

  size = generate(); 	/* Generates SIZE points */

  first = &hull[0];
  last = &hull[size - 1];

  /* Point with maximum distance to first-last*/
  max = maxDistancePoint( first, last);

  BARRIER;
  GETTIME(TimeP1);  
    par_q_hull( &first, &last, max, &size);
  GETTIME(TimeP2); 
  DIFTIME(TimeP2,TimeP1,t_par[REP]);
 

  /* show_output(first, last, size); */ 
  
 

  /*** REPORTE DE TIEMPOS  **************************************************/
	
	BARRIER;
  	if (NAME == 0) { 
	PRINTF("PARALLEL QSORT WRITTEN IN MPI-llc. ARGENTINA. RECURSIVE COMMON-COMMON VERSION     \n");
	PRINTF("SEQUENTIAL     \tREP            \tNAME           \tDEEP           \tSIZE           \tENDSEQ (P=1)      \tPAR            \tNUMPROC        \tSPEEDUP        \tDIVISION       \tCOMBIN         \tFIRST.ST.      \tCOMMUNICAT	\tSIZE-ENV	\tSIZE-REC	\n");
	}
 
  for(REP = 0; REP < MAX_REP; REP++)
	{
    for (name = 0; name <  NUMPROCESSORS; name++) {
		 if (NAME == name) {
			 for (d=0; d<deep1; d++) {
				 printf("%6.8lf     \t",t_seq[REP]);
				 printf("%6d    \t%6d     \t%6d     \t%6d      \t",REP, NAME, d,SIZE);
				 printf("%6.8lf     \t%6.8lf     \t", t_sequential[REP], t_par[REP]);
				 printf("%6d     \t",NUMPROCESSORS);
				 if  (t_par[REP] != (double) 0.0) printf("%6.8lf     \t", (double)(t_seq[REP]/t_par[REP]));
				 else printf("warning!!!!!!     \n");
				 printf("%6.8lf     \t%6.8lf     \t", t_division[REP][d], t_combination[REP][d]);
				 printf("%6.8lf      \t", t_fst[REP]);
				 printf("%6.8lf	\t", t_comm[REP][d]);
 				printf("%d	\t", sizeenv[REP][d]);
 				printf("%d\t", sizerec[REP][d]);

				 printf("     \n");
			 } /* for */
		 } /* if */
		   
		 fflush(stdout);
		 BARRIER;
    } /* for */
		 		 
		/* SIZE = 2*SIZE;*/
		 BARRIER;
  }

	 fflush(stdout); 
	 BARRIER;


  EXIT; 
  /*MPI_Finalize();*/
  exit(0);

} /* main */
